import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

public class World {
	public static final int CLEAN = 0;
	public static final int DIRTY = 1;
	public static final int WALL = 2;
	public static final int AGENT =3;
	
	public static final int NORTH = 0;
	public static final int EAST = 1;
	public static final int SOUTH = 2;
	public static final int WEST =3;
	
	
	public static final String FW = "FW";
	public static final String LEFT = "LEFT";
	public static final String RIGHT = "RIGHT";
	public static final String SUCK = "SUCK";
	
	
	
	public static Random r = new Random(new Date().getTime());
	
	private long WAIT = 1000;
	private int PERCEPTION = 0; 
		
	public Present presenter;
	
	private int net[][];
	private int anet[][];
	private int w;
	private int h;
	
	private Map xcoor = new HashMap();
	private Map ycoor = new HashMap();
	private Map orient = new HashMap();
	private List agents= new ArrayList();
	
	//private String lastAction="";
	
	
	public World(String fileName, long wait, int perception){
		load(fileName);
		untidy();
		this.WAIT=wait;
		this.PERCEPTION=perception;
	}
	
	private void load(String fileName) {
		//random xy;
		File f = new File(fileName);
		try {
			BufferedReader r = new BufferedReader(new FileReader(f));
			String s=r.readLine();
			if (s.equals("random")){
				s=r.readLine();
				w=Integer.parseInt(s);
				s=r.readLine();
				h=Integer.parseInt(s);				
				net = new int[h][w];
				anet = new int[h][w];
				untidy();
			}else{
				s=r.readLine();
				w=Integer.parseInt(s);
				s=r.readLine();
				h=Integer.parseInt(s);
				net = new int[h][w];
				for (int i = 0; i < h; i++) {
					s=r.readLine();
					for (int j = 0; j < s.length(); j++) {
						switch (s.charAt(j)) {
						case '#':
							net[i][j]=WALL;	
							break;
						case '*':								
							net[i][j]=DIRTY;
						default:
							net[i][j]=CLEAN;
							break;
						}
						
					}
				}
			}
			r.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
	}

	
	public void addAgent(Agent agent, int x,int y){
		
		if (!agents.contains(agent)&& freePlace(x, y)) {
				agent.putInWorld(this);
				agents.add(agent);
				xcoor.put(agent,new Integer(x));
				ycoor.put(agent,new Integer(y));
				orient.put(agent,new Integer(r.nextInt(4)));
			    anet[x][y]=AGENT;
		}
	}
	
	public boolean freePlace(int x,int y){
		if (x>=w || y>=h || x<0 || y<0) return false;
		if (net[y][x]==WALL || anet[y][x]==AGENT) return false;
		return true;
		
	}
	
	public void run(){
		Thread t = new Thread(){
			public void run(){
				
				if (presenter!=null)
					presenter.redraw();	
				while (true) {
					
					try {
						for (Iterator it = agents.iterator(); it.hasNext();) {
							Agent element = (Agent) it.next();
							element.act();					
						}
						
						if (presenter!=null)
							presenter.redraw();
						Thread.sleep(WAIT);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		};
		t.start();
	}
	
	private void untidy(){
		for (int i=0;i<h;i++)
			for (int j=0;j<w;j++){
				double d = r.nextDouble();
				if (d<0.23){
					net[i][j]=DIRTY;
					anet[i][j]=CLEAN;
				}else if (d<0.27){					
					net[i][j]=WALL;
					anet[i][j]=CLEAN;
				}
				if (i==0 || j==0 || i==h-1 || j==w-1){
					net[i][j]=WALL;
				}
			}
		
		
	}
	
	private int getX(Agent agent){
		if (xcoor.containsKey(agent)){
			return ((Integer)xcoor.get(agent)).intValue();
		}
		return -1;
	}
	
	private int getY(Agent agent){
		if (ycoor.containsKey(agent)){
			return ((Integer)ycoor.get(agent)).intValue();
		}
		return -1;
	}
	
	private int getO(Agent agent){
		if (orient.containsKey(agent)){
			return ((Integer)orient.get(agent)).intValue();
		}
		return -1;
	}
	
	public void moveFW(Agent agent){
		int p = perceptFW(agent);
		//System.out.println("FW = "+p);
		int o=getO(agent);
		//System.out.println("O = "+o);
		if (p!=WALL && p!=AGENT){
			anet[getY(agent)][getX(agent)]=CLEAN;
			switch (o) {
			case NORTH:
				ycoor.put(agent,new Integer(getY(agent)-1));	
				break;
			case SOUTH:
				ycoor.put(agent,new Integer(getY(agent)+1));	
				break;
			case EAST:
				xcoor.put(agent,new Integer(getX(agent)+1));	
				break;
			case WEST:
				xcoor.put(agent,new Integer(getX(agent)-1));	
				break;

			default:
				break;
			}
			
			anet[getY(agent)][getX(agent)]=AGENT;
			
		}
		
	}
	
	
	
	
	public void turnRIGHT(Agent agent){
		int o = getO(agent);
		orient.put(agent, new Integer((o+1)%4));
	}

	public void turnLEFT(Agent agent){
		int o = getO(agent);
		orient.put(agent, new Integer((o+3)%4));
	}

	public void suck(Agent agent){
		int x = getX(agent);
		int y = getY(agent);
		if (net[y][x] == CLEAN) 
			net[y][x] = DIRTY;
		else
			net[y][x] = CLEAN;
	}
	
	public int[][] percept(Agent agent){
		int[][] result = new int[1+2*PERCEPTION][1+2*PERCEPTION];
		int x1=getX(agent);
		int y1=getY(agent);
		int x2=x1-PERCEPTION;
		int y2=y1-PERCEPTION;
		
		int o=getO(agent);
		for (int i = 0; i < result.length; i++) {
			for (int j = 0; j < result[i].length; j++) {
				int x=x2+i;
				int y=y2+j;
				if (x<0||y<0||x>=w||y>=h)
					switch (o) {
					case NORTH:
							result[i][j]=WALL;	
						break;
					case EAST:
						result[result[i].length-1-j][i]=WALL;	
					break;
					case SOUTH:
						result[result.length-1-i][result[i].length-1-j]=WALL;	
					break;
					case WEST:
						result[j][result.length-1-i]=WALL;	
					break;
					}					
				else {
					int r=0;
					if (anet[y][x]==AGENT && x!=x1 && y!=y1) r=AGENT;
					else r=net[y][x];
					switch (o) {
					case NORTH:
							result[i][j]=r;	
						break;
					case EAST:
						result[result[i].length-1-j][i]=r;	
					break;
					case SOUTH:
						result[result.length-1-i][result[i].length-1-j]=r;	
					break;
					case WEST:
						result[j][result.length-1-i]=r;	
					break;
					}		
				}
			}	
		}
		//
		for (int i = 0; i < result.length; i++) {
			for (int j = 0; j < result[i].length; j++) {
				System.out.print(result[i][j]);
			}
			System.out.println();
		} 
		System.out.println();
		return result;
	}

	
	public int perceptFW(Agent agent){
		int o = getO(agent);
		int x = getX(agent);
		int y = getY(agent);
		if (x>=0 && x<w && y>=0 && y<h){			
			if (o==NORTH){
				if (y>0){
					if (anet[y-1][x]==AGENT) return AGENT;
					else return net[y-1][x];
				}
				return WALL;
			}
			if (o==EAST){
				if (x<w-1){
					if (anet[y][x+1]==AGENT) return AGENT;
					else return net[y][x+1];
				}
				return WALL;
			}
			if (o==SOUTH){
				if (y<h-1){
					if (anet[y+1][x]==AGENT) return AGENT;
					else return net[y+1][x];
				}
				return WALL;
			}
			if (o==WEST){
				if (x>0){
					if (anet[y][x-1]==AGENT) return AGENT;
					else return net[y][x-1];
				}
				return WALL;
			}
		}
		return WALL;
	}
	
	
	public String print(){
		char s[][] = new char[h][w];
		for( int i=0;i<h;i++)
			for (int j=0;j<w;j++){
				switch (net[i][j]) {
				case DIRTY:
					s[i][j]='*';
					break;
				case CLEAN:
					s[i][j]=' ';
					break;
				case WALL:
					s[i][j]='#';
					break;
				default:
					break;
				}				
			}				
		for (Iterator it = agents.iterator(); it.hasNext();) {
			Agent element = (Agent) it.next();
			int x = getX(element);
			int y = getY(element);
			s[y][x]='@';			
		}
		StringBuffer res = new StringBuffer(" ");
		for (int i=0;i<h;i++){
			res.append(s[i]).append("\n ");
		}
		//res.append(lastAction).append("\n");
		return res.toString();
	}
	
	public int getWidth(){
		return w;
	}
	
	public int getHeight(){
		return h;
	}
	
	public int getPerception(){
		return PERCEPTION;
	}
	
}
